print("import: 'skbio'")
import skbio

print("import: 'skbio.alignment'")
import skbio.alignment

print("import: 'skbio.alignment._lib'")
import skbio.alignment._lib

print("import: 'skbio.alignment.tests'")
import skbio.alignment.tests

print("import: 'skbio.diversity'")
import skbio.diversity

print("import: 'skbio.diversity.alpha'")
import skbio.diversity.alpha

print("import: 'skbio.diversity.alpha.tests'")
import skbio.diversity.alpha.tests

print("import: 'skbio.diversity.beta'")
import skbio.diversity.beta

print("import: 'skbio.diversity.beta.tests'")
import skbio.diversity.beta.tests

print("import: 'skbio.diversity.tests'")
import skbio.diversity.tests

print("import: 'skbio.io'")
import skbio.io

print("import: 'skbio.io.format'")
import skbio.io.format

print("import: 'skbio.io.format.tests'")
import skbio.io.format.tests

print("import: 'skbio.io.tests'")
import skbio.io.tests

print("import: 'skbio.sequence'")
import skbio.sequence

print("import: 'skbio.sequence.tests'")
import skbio.sequence.tests

print("import: 'skbio.stats'")
import skbio.stats

print("import: 'skbio.stats.distance'")
import skbio.stats.distance

print("import: 'skbio.stats.distance.tests'")
import skbio.stats.distance.tests

print("import: 'skbio.stats.evolve'")
import skbio.stats.evolve

print("import: 'skbio.stats.evolve.tests'")
import skbio.stats.evolve.tests

print("import: 'skbio.stats.ordination'")
import skbio.stats.ordination

print("import: 'skbio.stats.ordination.tests'")
import skbio.stats.ordination.tests

print("import: 'skbio.stats.tests'")
import skbio.stats.tests

print("import: 'skbio.tests'")
import skbio.tests

print("import: 'skbio.tree'")
import skbio.tree

print("import: 'skbio.tree.tests'")
import skbio.tree.tests

print("import: 'skbio.util'")
import skbio.util

print("import: 'skbio.util.tests'")
import skbio.util.tests

